@@ -50,10 +50,13 @@ way to do this is with your RDBMS's dump utility.  For example, for PostgreSQL,
     createdb littledb
     psql -f schemadump.sql littledb
 
-Currently rdbms-subsetter takes no account of schema names and simply assumes all
-tables live in the same schema.  This will probably cause horrible errors if used
-against databases where foreign keys span schemas.
+You can pull rows from a non-default schema by passing ``--schema=<name>``.
+Currently the target database must contain the corresponding tables in its own
+schema of the same name (moving between schemas of different names is not yet
+supported).
 
+Case-specific table names will probably create bad results in rdbms-subsetter,
+and in the rest of your life, for that matter.  Don't do it.
 """
 import argparse
 import logging
@@ -64,18 +67,19 @@ import types
 import sqlalchemy as sa
 from sqlalchemy.engine.reflection import Inspector
 
-__version__ = '0.2.0'
+__version__ = '0.2.1'
 
 def _find_n_rows(self, estimate=False):
     self.n_rows = 0
     if estimate:
         try:
             if self.db.engine.driver in ('psycopg2', 'pg8000',):
+                schema = (self.schema + '.') if self.schema else ''
                 qry = """SELECT reltuples FROM pg_class
-	                 WHERE oid = '%s'::regclass""" % self.name
+	                 WHERE oid = lower('%s%s')::regclass""" % (schema, self.name.lower())
             elif 'oracle' in self.db.engine.driver:
                 qry = """SELECT num_rows FROM all_tables
-	                 WHERE table_name='%s'""" % self.name
+	                 WHERE LOWER(table_name)='%s'""" % self.name.lower()
             else:
                 raise NotImplementedError("No approximation known for driver %s"
                                           % self.db.engine.driver)
@@ -86,6 +90,15 @@ def _find_n_rows(self, estimate=False):
     if not self.n_rows:
         self.n_rows = self.db.conn.execute(self.count()).fetchone()[0]
 
+def _random_row_func(self):
+    dialect = self.bind.engine.dialect.name
+    if 'mysql' in dialect:
+        return sa.sql.func.rand()
+    elif 'oracle' in dialect:
+        return sa.sql.func.dbms_random.value()
+    else:
+        return sa.sql.func.random()
+
 def _random_row_gen_fn(self):
     """
     Random sample of *approximate* size n
@@ -95,7 +108,7 @@ def _random_row_gen_fn(self):
             n = self.target.n_rows_desired
             if self.n_rows > 1000:
                 fraction = n / float(self.n_rows)
-                qry = sa.sql.select([self,]).where(sa.sql.functions.random() < fraction)
+                qry = sa.sql.select([self,]).where(self.random_row_func() < fraction)
                 results = self.db.conn.execute(qry).fetchall()
                 # we may stop wanting rows at any point, so shuffle them so as not to
                 # skew the sample toward those near the beginning
@@ -103,7 +116,7 @@ def _random_row_gen_fn(self):
                 for row in results:
                     yield row
             else:
-                qry = sa.sql.select([self,]).order_by(sa.sql.functions.random()).limit(n)
+                qry = sa.sql.select([self,]).order_by(self.random_row_func()).limit(n)
                 for row in self.db.conn.execute(qry):
                     yield row
 
@@ -130,7 +143,8 @@ def _pk_val(self, row):
         return None
 
 def _by_pk(self, pk):
-    pk_name = self.db.inspector.get_primary_keys(self.name)[0]
+    pk_name = self.db.inspector.get_primary_keys(self.name,
+                                                 self.schema)[0]
     slct = self.filtered_by(**({pk_name:pk}))
     return self.db.conn.execute(slct).fetchone()
 
@@ -140,49 +154,53 @@ def _exists(self, **kw):
 def _completeness_score(self):
     """Scores how close a target table is to being filled enough to quit"""
     result = ( 0
-              - (len(self.requested) / float(self.n_rows or 1 )) 
-              - len(self.required)) 
+              - (len(self.requested) / float(self.n_rows or 1 ))
+              - len(self.required))
     if not self.required:  # anything in `required` queue disqualifies
         result += (self.n_rows / (self.n_rows_desired or 1))**0.33
     return result
 
 class Db(object):
 
-    def __init__(self, sqla_conn, args):
+    def __init__(self, sqla_conn, args, schema=None):
         self.args = args
         self.sqla_conn = sqla_conn
+        self.schema = schema
         self.engine = sa.create_engine(sqla_conn)
-        self.meta = sa.MetaData(bind=self.engine)
-        self.meta.reflect()
+        self.meta = sa.MetaData(bind=self.engine) # excised schema=schema to prevent errors
+        self.meta.reflect(schema=self.schema)
         self.inspector = Inspector(bind=self.engine)
         self.conn = self.engine.connect()
         self.tables = OrderedDict()
         for tbl in self.meta.sorted_tables:
             tbl.db = self
+            # TODO: Replace all these monkeypatches with an instance assigment
             tbl.find_n_rows = types.MethodType(_find_n_rows, tbl)
-            tbl.fks = self.inspector.get_foreign_keys(tbl.name)
-            tbl.pk = self.inspector.get_primary_keys(tbl.name)
+            tbl.random_row_func = types.MethodType(_random_row_func, tbl)
+            tbl.fks = self.inspector.get_foreign_keys(tbl.name, schema=tbl.schema)
+            tbl.pk = self.inspector.get_primary_keys(tbl.name, schema=tbl.schema)
             tbl.filtered_by = types.MethodType(_filtered_by, tbl)
             tbl.by_pk = types.MethodType(_by_pk, tbl)
             tbl.pk_val = types.MethodType(_pk_val, tbl)
             tbl.exists = types.MethodType(_exists, tbl)
             tbl.child_fks = []
             tbl.find_n_rows(estimate=True)
-            self.tables[tbl.name] = tbl
-        for (tbl_name, tbl) in self.tables.items():
+            self.tables[(tbl.schema, tbl.name)] = tbl
+        for ((tbl_schema, tbl_name), tbl) in self.tables.items():
             for fk in tbl.fks:
-                fk['constrained_table'] = tbl_name
-                self.tables[fk['referred_table']].child_fks.append(fk)
+                fk['constrained_schema'] = tbl_schema
+                fk['constrained_table'] = tbl_name  # TODO: check against constrained_table
+                self.tables[(fk['referred_schema'], fk['referred_table'])].child_fks.append(fk)
 
     def __repr__(self):
         return "Db('%s')" % self.sqla_conn
 
     def assign_target(self, target_db):
-        for (tbl_name, tbl) in self.tables.items():
+        for ((tbl_schema, tbl_name), tbl) in self.tables.items():
             tbl._random_row_gen_fn = types.MethodType(_random_row_gen_fn, tbl)
             tbl.random_rows = tbl._random_row_gen_fn()
             tbl.next_row = types.MethodType(_next_row, tbl)
-            target = target_db.tables[tbl_name]
+            target = target_db.tables[(tbl_schema, tbl_name)]
             target.requested = deque()
             target.required = deque()
             if tbl.n_rows:
@@ -199,9 +217,13 @@ class Db(object):
             logging.debug("assigned methods to %s" % target.name)
 
     def confirm(self):
-        for tbl_name in sorted(self.tables):
-            tbl = self.tables[tbl_name]
-            print("Create %d rows from %d in %s" % (tbl.target.n_rows_desired, tbl.n_rows, tbl_name))
+        message = []
+        for (tbl_schema, tbl_name) in sorted(self.tables, key=lambda t: t[1]):
+            tbl = self.tables[(tbl_schema, tbl_name)]
+            message.append("Create %d rows from %d in %s.%s" %
+                           (tbl.target.n_rows_desired, tbl.n_rows,
+                            tbl_schema or '', tbl_name))
+        print("\n".join(sorted(message)))
         if self.args.yes:
             return True
         response = input("Proceed? (Y/n) ").strip().lower()
@@ -212,32 +234,35 @@ class Db(object):
         logging.debug('create_row_in %s:%s ' %
                       (target.name, target.pk_val(source_row)))
 
-        if target.exists(**(dict(source_row))):
-            logging.debug("Row already exists; not creating")
+        row_exists = target.exists(**(dict(source_row)))
+        logging.debug("Row exists? %s" % str(row_exists))
+        if row_exists and not prioritized:
             return
 
+        if not row_exists:
             # make sure that all required rows are in parent table(s)
-        for fk in target.fks:
-            target_parent = target_db.tables[fk['referred_table']]
-            slct = sa.sql.select([target_parent,])
-            any_non_null_key_columns = False
-            for (parent_col, child_col) in zip(fk['referred_columns'],
-                                               fk['constrained_columns']):
-                slct = slct.where(target_parent.c[parent_col] ==
-                                  source_row[child_col])
-                if source_row[child_col] is not None:
-                    any_non_null_key_columns = True
-            if any_non_null_key_columns:
-                target_parent_row = target_db.conn.execute(slct).first()
-                if not target_parent_row:
-                    source_parent_row = self.conn.execute(slct).first()
-                    self.create_row_in(source_parent_row, target_db, target_parent)
-        ins = target.insert().values(**source_row)
-        target_db.conn.execute(ins)
-        target.n_rows += 1
+            for fk in target.fks:
+                target_parent = target_db.tables[(fk['referred_schema'], fk['referred_table'])]
+                slct = sa.sql.select([target_parent,])
+                any_non_null_key_columns = False
+                for (parent_col, child_col) in zip(fk['referred_columns'],
+                                                   fk['constrained_columns']):
+                    slct = slct.where(target_parent.c[parent_col] ==
+                                      source_row[child_col])
+                    if source_row[child_col] is not None:
+                        any_non_null_key_columns = True
+                if any_non_null_key_columns:
+                    target_parent_row = target_db.conn.execute(slct).first()
+                    if not target_parent_row:
+                        source_parent_row = self.conn.execute(slct).first()
+                        self.create_row_in(source_parent_row, target_db, target_parent)
+
+            ins = target.insert().values(**source_row)
+            target_db.conn.execute(ins)
+            target.n_rows += 1
 
         for child_fk in target.child_fks:
-            child = self.tables[child_fk['constrained_table']]
+            child = self.tables[(child_fk['constrained_schema'], child_fk['constrained_table'])]
             slct = sa.sql.select([child,])
             for (child_col, this_col) in zip(child_fk['constrained_columns'],
                                              child_fk['referred_columns']):
@@ -255,17 +280,21 @@ class Db(object):
     def create_subset_in(self, target_db):
 
         for (tbl_name, pks) in self.args.force_rows.items():
-            source = self.tables[tbl_name]
+            if '.' in tbl_name:
+                (tbl_schema, tbl_name) = tbl_name.split('.', 1)
+            else:
+                tbl_schema = None
+            source = self.tables[(tbl_schema, tbl_name)]
             for pk in pks:
                 source_row = source.by_pk(pk)
                 if source_row:
-                    self.create_row_in(source_row, target_db, 
+                    self.create_row_in(source_row, target_db,
                                        source.target, prioritized=True)
                 else:
                     logging.warn("requested %s:%s not found in source db,"
                                  "could not create" % (source.name, pk))
 
-                    
+
         while True:
             targets = sorted(target_db.tables.values(),
                              key=lambda t: t.completeness_score())
@@ -285,13 +314,13 @@ class Db(object):
             if target.completeness_score() > 0.97:
                 return
             (source_row, prioritized) = target.source.next_row()
-            self.create_row_in(source_row, target_db, target, 
+            self.create_row_in(source_row, target_db, target,
                                prioritized=prioritized)
 
 
 def fraction(n):
     n = float(n)
-    if 0 < n <= 1:
+    if 0 <= n <= 1:
         return n
     raise argparse.ArgumentError('Fraction must be greater than 0 and no greater than 1')
 
@@ -312,7 +341,7 @@ argparser.add_argument('dest', help='SQLAlchemy connection string for data desti
                        type=str)
 argparser.add_argument('fraction', help='Proportion of rows to create in dest (0.0 to 1.0)',
                        type=fraction)
-argparser.add_argument('-l', '--logarithmic', help='Cut row numbers logarithmically; use 0.5 for fraction',
+argparser.add_argument('-l', '--logarithmic', help='Cut row numbers logarithmically; try 0.5 for fraction',
                        action='store_true')
 argparser.add_argument('--loglevel', type=loglevel, help='log level (%s)' % all_loglevels,
                        default='INFO')
@@ -321,6 +350,8 @@ argparser.add_argument('-f', '--force', help='<table name>:<primary_key_val> to
 argparser.add_argument('-c', '--children',
                        help='Max number of child rows to attempt to pull for each parent row',
                        type=int, default=3)
+argparser.add_argument('--schema', help='Non-default schema to include',
+                       type=str, action='append')
 argparser.add_argument('-y', '--yes', help='Proceed without stopping for confirmation', action='store_true')
 
 def generate():
@@ -332,8 +363,10 @@ def generate():
             args.force_rows[table_name] = []
         args.force_rows[table_name].append(pk)
     logging.getLogger().setLevel(args.loglevel)
-    source = Db(args.source, args)
-    target = Db(args.dest, args)
-    source.assign_target(target)
-    if source.confirm():
-        source.create_subset_in(target)
+    schemas = args.schema + [None,]
+    for schema in schemas:
+        source = Db(args.source, args, schema=schema)
+        target = Db(args.dest, args, schema=schema)
+        source.assign_target(target)
+        if source.confirm():
+            source.create_subset_in(target)
